# -*- coding: UTF-8 -*-
###########################################################################
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# util.py
#
# fonctions utilitaires pour la librairie sso eole
# (actuellement seulement annuaire LDAP)
#
###########################################################################
"""
    Librairies d'utilitaires pour le serveur eole-sso
"""
import urlparse, socket, re, os, IPy
import time, random, hashlib
from ConfigParser import ConfigParser

from twisted.python import log
from twisted.web2.http_headers import parseCookie

def urljoin(url, params):
    """ajoute des paramètres à une url
    >>> urljoin('http://test/','a=b')
    'http://test/?a=b'
    >>> urljoin('http://test/?a=b','c=d')
    'http://test/?a=b&c=d'
    """
    urllist = list(urlparse.urlsplit(url))
    if urllist[-2] == '':
        urllist[-2] = params
    else:
        urllist[-2] += '&' + params
    return urlparse.urlunsplit(urllist)

def format_err(msg):
    """
        renvoie une balise contenant une erreur
    """
    return """<h1><font color="red">%s</font></H1>""" % msg

def cas_response(resptype, body, **attrs):
    """
        convenience cas response factory
    """
    attrs = ' '.join('%s="%s"' % kv for kv in attrs.items())
    if attrs:
        attrs = ' ' + attrs
    return str("""<?xml version="1.0" encoding="UTF-8"?>
<cas:serviceResponse xmlns:cas="http://www.yale.edu/tp/cas">
  <cas:%(rtype)s%(attrs)s>
    %(body)s
  </cas:%(rtype)s>
</cas:serviceResponse>\r\n\r\n""" % \
{'rtype': resptype, 'attrs': attrs, 'body': body})

def unescape(url, quote=None):
    """
        supprime les caractères escapés dans une url
    """
    entities = [('&gt;','>'), ('&lt;','<'), ('&amp;','&')]
    if quote:
        entities.insert(0, ('&quot;','"'))
    for substr, car in entities:
        url = url.replace(substr, car)
    return url

def get_service_from_url(url):
    """
        supprime la partie requête d'une url
        pour permettre d'indentifier le service
    """
    url = unescape(url)
    u = urlparse.urlparse(url)
    service = "%s://%s%s" % (u.scheme, u.netloc, u.path)
    return service

def getCookies(ctx):
    """
        renvoie la liste  des cookies sous forme Twisted
    """
    cookiestr = ctx.headers.getRawHeaders('Cookie')
    if cookiestr:
        return parseCookie(cookiestr)
    return []

def getCookie(ctx, key):
    """renvoie un cookie Twisted"""
    for cookie in getCookies(ctx):
        if cookie.name == key:
            return cookie
    return None

def verif_cert(conn, cert, err_num, err_depth, rc):
    """fonction de vérification du certificat"""
    return not cert.has_expired()

def check_hostname_by_ip(url, addr):
    """vérification de la validité de l'hôte de l'url en fonction d'une IP
    """
    try:
        # vérification de l'adresse
        ip = IPy.IP(addr)
        try:
            hostname = socket.gethostbyname(url.hostname)
        except:
            hostname = url.hostname
        assert ip.overlaps(hostname)
        return True
    except:
        return False
    return False

def check_hostname_by_domain(url, addr):
    """
        vérification de la validité de l'hôte de l'url
        par expression régulière
    """
    #try:
    #    hostname = socket.gethostbyaddr(url.hostname)[0]
    #except:
    hostname = url.hostname
    # on essaye un match de l'expression régulière sur le nom d'hôte
    try:
        match_obj = re.match(addr, hostname)
    except:
        log.msg(_("""Error matching regexp '%s' (invalid syntax ?)""") % addr)
        return False

    return match_obj is not None

def gen_random_id(prefix=''):
    """
        Renvoie un identifiant aléatoire
        :prefix: préfixe à placer devant la string
    """
    random_id = str(time.time())
    return prefix + hashlib.sha224('%s/%s' % (
                    random_id, random.randrange(2**100))).hexdigest()

def gen_ticket_id(prefix, address):
    """
        à redéfinir dans la classe dérivée
        génère un id de ticket ex : ST-monetab.ac-acad.fr-51de51d5e1d5e1d51d5e1d5e1
        :prefix: le préfixe décrivant le type de ticket
        :address: l'adresse associée au ticket fourni
    """
    ticket_id = prefix + '-' + address + '-' + gen_random_id()
    return ticket_id

def get_replication_branches():
    """
        returns search_bases as defined in /etc/ldap/replication.conf
        if available
    """
    F_CONF_REPL = '/etc/ldap/replication.conf'
    search_bases = {}
    libel_etab = ""
    if os.path.isfile(F_CONF_REPL):
        f_repl = file(F_CONF_REPL)
        for line in f_repl:
            line = line.strip()
            if line.startswith('#') and line.endswith(')'):
                # on récupère le libellé de l'établissement qui
                # apparait en début de section
                libel_etab = line[1:].strip()
            if line.startswith("searchbase="):
                branch = line.strip()
                # on supprime searchbase= et les guillemets
                branch = branch[branch.index("=") + 2:-1]
                if not libel_etab:
                    # si libellé non trouvé, on cherche le rne
                    # (normalement, premier objet du dn de recherche)
                    libel_etab = branch.split(',')[0]
                    libel_etab = libel_etab[libel_etab.index('=') + 1:]
                search_bases[branch] = libel_etab
                libel_etab = ""
        f_repl.close()
    return search_bases

class EoleParser(ConfigParser):
    """Sous classe de ConfigParser pour préserver la casse des options"""

    def optionxform(self, option_str):
        return option_str
