#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2011
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# check_certs.py
#
# procédure de vérification de la validité des certificats
#
# compatible avec les versions 1 et 2 de la librairie CAS
# (http://www.ja-sig.org/products/cas/)
#
###########################################################################

# Imports pour la gestion de SSL via M2Crypto
from M2Crypto import SSL, X509
# Imports EoleSSO
from eolesso.libsecure import ClientContextFactory
import config, socket, sys

try:
    from creole.utils import print_orange
except:
    # librairie Creole non disponible, on ne vérifie pas le certificat
    sys.exit(0)

missing_msg = """

Attention, les adresses suivantes ne sont pas définies
comme sujet du certificat %s : %s
"""

warn_msg = """
Les applications utilisant un client CAS en mode proxy
(webmail, ftp, ...) risquent de ne pas fonctionner correctement.
En cas d'utilisation du certificat par défaut (/etc/ssl/eole.crt),
vous pouvez le regénérer avec la commande /usr/share/creole/gen_certif.py -f
"""

def run_check():
    # initialisation d'un contexte client permettant de vérifier la validité
    # des certificat (mode proxy)
    client_ctx = ClientContextFactory(certfile = config.CERTFILE, keyfile = config.KEYFILE,
                               mode = SSL.m2.SSL_VERIFY_PEER|SSL.m2.SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
                               ca_location = config.CA_LOCATION)

    peer_cert = X509.load_cert(config.CERTFILE)
    res_verify = client_ctx._cert_verify(peer_cert)
    addr_err = []
    if res_verify:
        # on détermine les adresses à vérifier
        test_addrs = [config.SERVER_IP_ADDR]
        # on vérifie toujours la connexion sur eth0
        local_fqdn = socket.getfqdn(config.SERVER_IP_ADDR)
        if config.AUTH_SERVER_ADDR != local_fqdn:
            # domaine particulier défini (cas_domainname ?)
            test_addrs.append(config.AUTH_SERVER_ADDR)
        # on vérifie les adresses présentes dans le certificat du serveur
        # subject + subjectAltnames
        for addr in test_addrs:
            if not addr in client_ctx.peer_cert_data['subject']:
                addr_err.append(addr)
                res_verify = False
    if len(addr_err) > 0:
        print_orange(missing_msg % (config.CERTFILE, ", ".join(addr_err)))
        print warn_msg
    return res_verify

if __name__ == '__main__':
    run_check()
