#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# saml_idp.py
#
# support (limité) du protocole SAML 2.0 pour eole-sso
#
###########################################################################

import traceback
import urllib, os, socket
from cgi import escape

from twisted.web2 import http, responsecode
from twisted.web2.http_headers import Cookie as TwCookie, MimeType
from twisted.web2.resource import PostableResource
from twisted.internet.threads import callMultipleInThread
from twisted.python import failure

from pyeole import httprequest
from eolesso.util import (getCookies, gen_random_id, gen_ticket_id,
                          format_err, get_service_from_url)

import saml_response
from saml_crypto import sign_request
from saml_utils import (encode_request, gen_metadata, get_endpoint,
                        get_attributes, InternalError, Redirect,
                        extract_message, check_required_contexts,
                        process_auth_request, process_assertion)
from saml2 import samlp
from page import gen_page, trace, log
from config import (CERTFILE, AUTH_FORM_URL, IDP_IDENTITY, DEFAULT_CSS,
                    SEND_CAS_LOGOUT, DISPLAY_FEDERATION, encoding)

class SamlResource(PostableResource):
    content_type = 'text/html'

    def render(self, request):
        headers = {'Content-type': MimeType.fromString(self.content_type)}
        # errheaders = {'Content-type': MimeType.fromString('application/xml')}
        try:
            return self._render(request)
        except Redirect, exc:
            location = exc.location
            headers['location'] =  location
            log.msg(_('Redirecting to '), location)
            # 303 See other
            return http.Response(code = 303, headers = headers)
        except Exception, exc: # unknown error
            traceback.print_exc()
            return http.StatusResponse(code = http.responsecode.INTERNAL_SERVER_ERROR, description = str(exc))
            # return http.Response(stream = error.as_xml(), headers = errheaders)

    def _render(self):
        pass

def gen_page_err(err_msg, from_url, css):
    """construit une réponse d'erreur
    """
    form = format_err(err_msg)
    if from_url:
        form += """<p><FORM action="%s">
        <b>%s</b> : <br/><h2><font color="blue">%s</font></h2><br/>
        <input class="btn" type="submit" value ="%s"/>
        </form></p>""" % (from_url, _('Following page has been proposed'), from_url, _('Proceed to above url'))
    else:
        form += """<br/><form action="/login" method="post">
        <input class="btn" type="submit" value="%s"></form>""" % _('New session')
    return http.Response(stream = gen_page(_('Logout'), form, css))

class SamlLogout(SamlResource):
    """Resource de gestion du logout (logout cas + single logout saml)
    """
    addSlash = False

    def __init__(self, manager):
        self.manager = manager
        super(SamlLogout, self).__init__()

    @trace
    def send_logout_request(self, sp_identity, sp_meta, user_id, ticket, sso_session, from_url, css):
        """Envoi d'une requête de déconnexion à une entité partenaire et stocke les informations sur l'origine de la déconnexion"""
        try:
            binding, service_url, response_url = get_endpoint(sp_meta, 'SingleLogoutService')
        except InternalError:
            # store error for display ?
            return False

        response_id = gen_random_id('_')
        if hasattr(ticket, 'idp_session_index'):
            session_index = ticket.idp_session_index
        else:
            session_index = ticket.ticket
        # Envoi de la demande de logout à l'entité correspondante
        if binding == samlp.BINDING_HTTP_POST:
            sign_method, req = saml_response.gen_logout_request(response_id, session_index, user_id, IDP_IDENTITY, \
                    sp_meta['entityID'], service_url, CERTFILE)
            # réponse utilisant le binding POST
            req = encode_request(req)
            form = """
                    <FORM action="%s" method="POST">
                    <input type="hidden" name="SAMLRequest" value="%s">
                    <input type="hidden" name="SigAlg" value="%s">
                    <input type="hidden" name="RelayState" value="%s">
                    <input class="btn" type="submit" value="%s">
            """ % (service_url, req, sign_method, from_url, _("Proceed to service"))
            resp = http.Response(stream = gen_page(_('Logout request to service'), form, css))
        else:
            sign_method, req = saml_response.gen_logout_request(response_id, session_index, user_id, IDP_IDENTITY, \
                    sp_meta['entityID'], service_url, CERTFILE, sign = False)
            # réponse utilisant le binding REDIRECT
            req = encode_request(req, True)
            req_args = sign_request(CERTFILE, sign_method, 'SAMLRequest', req, from_url)
            if req_args != []:
                redirect_url = '%s?%s' % (service_url, urllib.urlencode(req_args))
            resp = http.RedirectResponse(redirect_url)

        # on stocke l'id et le ticket avant de rediriger
        ticket.valid = False
        self.manager.pending_logout[sso_session][3][response_id] = ticket.ticket
        log.msg(_("Sending SAML logout request to %s (%s) : %s") % (sp_identity, service_url, response_id))
        return resp

    @trace
    def send_cas_logout_request(self, ticket, use_SOAP = False):
        """
            envoie une demande de déconnexion à un client CAS par un requête SAML
            use_SOAP : utiliser une enveloppe SOAP pour transmettre le message si True
                       (versions futures du client CAS ?)
        """
        logoutRequest = saml_response.gen_cas_logout_request(ticket.ticket,
                                                                  use_SOAP)
        log.msg(_("Sending CAS logout request to %s") % ticket.service_url)
        # lancement de la requête
        req = httprequest.HTTPRequest()
        try:
            req.request(ticket.service_url, {'logoutRequest':logoutRequest})
        except httprequest.RequestError, e:
            log.msg(_('Error sending logout request to %s : ' % (
                                                      ticket.service_url,)))
            log.msg(str(e))

    @trace
    def process_logout_response(self, request, sso_session):
        """traite une réponse de déconnexion et met à jour les tickets correspondants dans le gestionnaire de sessions
        """
        signature_ok, saml_resp = extract_message(request, 'SAMLResponse')
        if not signature_ok:
            # FIXME TODO: generate a saml negative saml response
            raise InternalError("%s : %s" % (_('signature error'), str(saml_resp)))
        response = samlp.LogoutResponseFromString(saml_resp)
        resp_issuer = response.issuer.text.encode(encoding)
        in_response_to = response.in_response_to
        # Vérification de l'existence de la requête spécifiée dans in_response_to
        if in_response_to not in self.manager.pending_logout[sso_session][3]:
            # FIXME TODO: generate a saml negative saml response
            raise InternalError("%s : %s" % (_('corresponding logout request not found'), in_response_to))
        #issue_instant = date_from_string(response.issue_instant)
        status_code = response.status.status_code.value
        #if response.status.status_message is not None:
        #    status_msg = response.status.status_message.text
        ticket_id = self.manager.pending_logout[sso_session][3][in_response_to]
        if status_code == samlp.STATUS_SUCCESS:
            self.manager.pending_logout[sso_session][1].append(ticket_id)
        log.msg(_('SAMLResponse (%s) received from %s in response to %s') % ('LogoutResponse', resp_issuer, in_response_to))

    @trace
    def process_logout_request(self, request, sso_session):
        """traite une demande de déconnexion et initie le processus de déconnexion
        """
        signature_ok, saml_req = extract_message(request, 'SAMLRequest')
        if not signature_ok:
            # FIXME TODO: generate a saml negative saml response
            raise InternalError("signature error : %s" % str(saml_req))
        saml_req = samlp.LogoutRequestFromString(saml_req)
        req_issuer = saml_req.issuer.text.encode(encoding)
        req_id = saml_req.id
        app_ticket = saml_req.session_index.text.encode(encoding)
        if app_ticket in self.manager.sp_sessions:
            # session définie par l'idp distant -> on récupère l'app_ticket local correspondant
            app_ticket = self.manager.sp_sessions[app_ticket]
        if app_ticket in self.manager.app_sessions:
            ticket = self.manager.app_sessions[app_ticket]
            if ticket.session_id == sso_session:
                if req_issuer != ticket.sp_ident:
                    # FIXME TODO: generate a saml negative saml response
                    raise InternalError(_('logout requester (%s) does not own session') % req_issuer)
                # début d'un logout initié par un participant de la session
                self.manager.pending_logout[sso_session] = [samlp.STATUS_SUCCESS, [app_ticket], (app_ticket, req_id, req_issuer), {}, []]
                # on défini le ticket comme non valid
                ticket.valid = False
            else:
                # FIXME TODO: generate a saml negative saml response
                raise InternalError(_('logout request for invalid session'))
        else:
            # FIXME TODO: generate a saml negative saml response
            raise InternalError(_('Invalid session'))
        log.msg(_('SAMLRequest (%s) received from %s') % ('LogoutRequest', req_issuer))

    @trace
    def check_attached_sessions(self, sso_session, closed_services, failed_services, from_url, css):
        """Vérifie les tickets rattachés à la session en cours et envoie les demandes de déconnexion nécessaires
        maintient la liste échecs/succés de chaque demande
        """
        # recherche des sessions non traitées lors du logout
        self.manager.pending_logout[sso_session][0] = samlp.STATUS_SUCCESS
        services_done = self.manager.pending_logout[sso_session][4]
        if from_url not in services_done:
            services_done.append(from_url)
        logout_call_list = []
        # on regarde si le logout a été initié par requête d'une entité partenaire
        if self.manager.pending_logout[sso_session][2]:
            req_ticket, req_id, req_issuer  = self.manager.pending_logout[sso_session][2]
        else:
            req_ticket = req_id = req_issuer = None

        for ticket in self.manager.user_app_tickets[sso_session]:
            if hasattr(ticket,'sp_ident'):
                if ticket.ticket != req_ticket:
                    if ticket.valid:
                        # ticket correspondant à une session 'saml' valide
                        # -> Envoi d'une demande de déconnexion auprès de l'entité partenaire
                        sp_identity = ticket.sp_ident
                        # Lecture des métadonnées du partenaire
                        sp_meta = self.manager.get_metadata(sp_identity)
                        user_id = ticket.name_id
                        return self.send_logout_request(sp_identity, sp_meta, user_id, ticket, sso_session, from_url, css)
                    else:
                        # Une déconnexion a déjà été demandée pour ce ticket
                        # -> On met à jour le status du logout en cours
                        if ticket.ticket in self.manager.pending_logout[sso_session][1]:
                            closed_services.append(ticket.sp_ident)
                        else:
                            failed_services.append(ticket.sp_ident)
                            # Au moins une des demandes envoyées n'a pas reçu de réponse favorable (logout partiel)
                            self.manager.pending_logout[sso_session][0] = samlp.STATUS_PARTIAL_LOGOUT
            else:
                if SEND_CAS_LOGOUT:
                    # ticket CAS standard, on envoie une requête de logout
                    # au client CAS (sur l'url de service du ticket).
                    if ticket.service_url not in services_done:
                        services_done.append(ticket.service_url)
                        if not ticket.ticket.startswith('PT-'):
                            # on n'envoie pas de demande de déconnexion
                            # sur les services authentifiés en mode proxy (imap, ftp, ...)
                            logout_call = (self.send_cas_logout_request,
                                        [ticket, False],
                                        {})
                            logout_call_list.append(logout_call)
        if logout_call_list:
            # On appelle l'ensemble des logouts cas dans un thread
            callMultipleInThread(logout_call_list)
        return None


    @trace
    def build_logout_response(self, sso_session, relay_state, css):
        """construit une réponse à destination de l'entité ayant demandé la déconnexion
        """
        # Lecture des informations sur l'initiateur de la déconnexion
        app_ticket, req_id, req_issuer  = self.manager.pending_logout[sso_session][2]
        # Si le logout est effectué suite à une LogoutRequest, on envoie une réponse au demandeur
        sp_meta = self.manager.get_metadata(req_issuer)
        try:
            binding, service_url, response_url = get_endpoint(sp_meta, 'SingleLogoutService')
        except InternalError:
            traceback.print_exc()
            #XXX FIXME
            err_msg = _('no usable endpoint for %s') % req_issuer
            return http.Response(stream = gen_page(_('Logout Error'), err_msg, css))
        # Envoi de la réponse finale à l'émetteur de la requête
        response_id = gen_ticket_id('SAMLResponse', self.manager.address)
        sign_method, response = saml_response.gen_logout_response(req_id, response_id, IDP_IDENTITY, response_url, CERTFILE, \
                self.manager.pending_logout[sso_session][0])
        log.msg(_("SAMLResponse (%s) sent to %s") % ('LogoutResponse', response_url))
        if binding == samlp.BINDING_HTTP_REDIRECT:
            # Réponse avec le binding REDIRECT
            response = encode_request(response, True)
            # la signature doit se faire sur la chaine suivante : SAMLResponse=value&RelayState=value&SigAlg=value -> encodé dans location header
            req_args = sign_request(CERTFILE, sign_method, 'SAMLResponse', response, relay_state)
            if req_args != []:
                redirect_url = '%s?%s' % (response_url, urllib.urlencode(req_args))
            else:
                redirect_url = response_url
            resp = http.RedirectResponse(redirect_url)
        else:
            # Réponse avec le binding POST (default fallback)
            response = encode_request(response)
            relay_info = ""
            if relay_state:
                relay_info = """<input type="hidden" name="RelayState" value="%s">""" % relay_state
            form = """
                    <FORM action="%s">
                    <input type="hidden" name="SAMLResponse" value="%s">
                    %s
                    <input type="hidden" name="SigAlg" value="%s">
                    <input class="btn" type="submit" value="%s">
            """ % (relay_state, response_url, response, relay_info, sign_method, _('Proceed to service'))
            resp = http.Response(stream = gen_page(_('Logout confirmation to service'), form, css))
        return resp

    @trace
    def build_logout_page(self, sso_session, closed_services, failed_services, from_url, css):
        """construit la page à afficher en cas de déconnexion initiée localement
        """
        logout_msg = ""
        if closed_services != []:
            logout_msg += """<b><font color="green">%s</font></b>""" % _("Following service providers have been disconnected")
            logout_msg += "<br/><table border=0><tr><td>%s</td></tr></table>" % "</td></tr><tr><td>".join(closed_services)
        if failed_services != []:
            logout_msg += """<b><font color="orange">%s</font></b>""" % _("Following service providers did not report successful logout")
            logout_msg += "<br/><table border=0><tr><td>%s</td></tr></table>" % "</td></tr><tr><td>".join(failed_services)
        if from_url:
            logout_msg += """<form action="%s">
            <br/><b>%s</b> : <br/><h2><font color="blue">%s</font></h2><br/>
            <input class="btn" type="submit" value ="%s"/>
            </form></p>""" % (from_url, _('Following page has been proposed'), from_url, _('Proceed to above url'))
        else:
            logout_msg += """<br/><form action="/login" method="POST">
            <input class="btn" type="submit" value="%s"></form>""" % _('New session')
        if self.manager.pending_logout[sso_session][0] == samlp.STATUS_PARTIAL_LOGOUT:
            logout_status = _("Partial single logout success")
        else:
            logout_status = _("Single logout success")
        return http.Response(stream = gen_page(logout_status, str(logout_msg), css))

    def _render(self, request):
        """Gestion de la déconnexion des sessionss SSO
        """
        # vérification de la session en cours
        cookie = None
        authenticated = False
        cookies = getCookies(request)
        sso_session = None
        for cookie in cookies:
            if cookie.name == 'EoleSSOServer':
                sso_session = cookie.value
                authenticated = self.manager.validate_session(sso_session)
                break

        css = DEFAULT_CSS
        # vérifie si on doit rediriger après la déconnexion
        from_url = None
        relay_state = None
        if 'RelayState' in request.args:
            relay_state = request.args['RelayState'][0]
            from_url = relay_state
        elif request.args.has_key('service'):
            from_url = request.args['service'][0]
        elif request.args.has_key('url'):
            from_url = request.args['url'][0]
        # recherche d'une eventuelle css associée
        if from_url:
            service_filter = self.manager._check_filter(from_url)
            if service_filter not in ['default', '']:
                if os.path.exists(os.path.join('interface', '%s.css' % service_filter)):
                    css = service_filter
        try:
            css = request.args['css'][0]
        except KeyError:
            pass
        closed_services = []
        failed_services = []
        resp = None

        try:
            ########################
            # Gestion du logout SAML
            if not authenticated:
                # pas d'authentification détectée
                raise InternalError(_('Invalid session'))
            else:
                # traitement d'une requête reçue : LogoutRequest ou LogoutResponse
                if 'SAMLResponse' in request.args:
                    self.process_logout_response(request, sso_session)
                if 'SAMLRequest' in request.args:
                    self.process_logout_request(request, sso_session)
                elif sso_session not in self.manager.pending_logout:
                    # Logout initié localement (idp initiated)
                    self.manager.pending_logout[sso_session] = [samlp.STATUS_SUCCESS, [], None, {}, []]
                # envoi de requêtes de déconnexion pour chaque session attachée à la session SSO
                resp = self.check_attached_sessions(sso_session, closed_services, failed_services, from_url, css)
                if resp is not None:
                    return resp
                # Tous les participants à la session sont prévenus, calcul du résultat final
                if self.manager.pending_logout[sso_session][2]:
                    # Envoi d'une réponse à la demande de déconnexion
                    resp = self.build_logout_response(sso_session, relay_state, css)
                else:
                    # affichage d'un message (déconnexion locale)
                    resp = self.build_logout_page(sso_session, closed_services, failed_services, from_url, css)

            ############################
            # Logout SSO Session Globale

            # suppression finale de la session sso et des données associées
            self.manager.logout(sso_session)
        except InternalError, err:
            log.msg(err)

        if from_url and failed_services == []:
            # Dans le cas ou on doit rediriger et aucune erreur n'est survenue -> pas d'affichage
            # XXX FIXME : display a warning form in any case ?
            resp = http.RedirectResponse(from_url)
            log.msg(_("Logged out : redirecting to %s") % from_url)
        elif resp is None:
            resp = http.RedirectResponse('/')

        # expiration du cookie EoleSSO
        if sso_session and cookie:
            cookie.expires = 1
            resp.headers.setHeader('Set-Cookie', cookies)
        return resp

class SamlMetadata(SamlResource):

    def __init__(self, manager):
        self.manager = manager
        super(SamlMetadata, self).__init__()

    def _render(self, request):
        """Affiche les métadonnées SAML du serveur local"""
        data = gen_metadata(self.manager, AUTH_FORM_URL, CERTFILE)
        resp = http.Response(stream = data)
        return resp

class SamlErrorPage(SamlResource):
    """
    page d'erreur http
    """
    addSlash = False

    def __init__(self, code, description):
        self.code = code
        self.description = description
        title = ('Eole-SSO : Erreur')
        super(SamlErrorPage, self).__init__()

    def _render(self, request):
        return http.StatusResponse(code = self.code, description = self.description)

class SamlConsumer(SamlResource):
    """Resource de traitement des demandes d'accès aux service tiers
    (Fournisseur de service)
    """
    addSlash = False

    def __init__(self, manager):
        self.manager = manager
        super(SamlConsumer, self).__init__()

    @trace
    def remove_old_ticket(self, session_id, sp_ident):
        """ deletes any existing session for a given identity provider
        """
        for ticket in self.manager.user_app_tickets[session_id]:
            if getattr(ticket, 'sp_ident', '') == sp_ident:
                old_ticket = ticket.ticket
                # suppression du ticket des logout en cours si i len fait partie
                if session_id in self.manager.pending_logout:
                    if old_ticket in self.manager.pending_logout[session_id][1]:
                        self.manager.pending_logout[session_id][1].remove(old_ticket)
                # suppression du ticket de la session utilisateur
                self.manager.user_app_tickets[session_id].remove(ticket)
                # suppression des éventuelles indexs de session distants associés (authentification à travers un deuxième idp)
                if hasattr(ticket, 'idp_session_index'):
                    if ticket.idp_session_index in self.manager.sp_sessions:
                        del(self.manager.sp_sessions[ticket.idp_session_index])
                # suppression du ticket d'application
                del(self.manager.app_sessions[old_ticket])
                del(ticket)
                break

    def _render(self, request):
        """Assertion Consuming processing
        """
        # vérification d'une éventuelle session en cours
        authenticated = False
        cookies = getCookies(request)
        sso_session = None
        for cookie in cookies:
            if cookie.name == 'EoleSSOServer':
                sso_session = cookie.value
                # On vérifie si la session est valide
                authenticated = self.manager.validate_session(sso_session)
                break
        # XXX FIXME : si déjà authentifié: supprimer le cookie actuel ou vérifier qu'il correspond au bon utilisateur ?
        # pb: single logout de l'ancienne session ?
        css = DEFAULT_CSS
        # vérification de l'adresse où l'utilisateur doit être redirigé
        relay_state = None
        if 'RelayState' in request.args:
            relay_state = request.args['RelayState'][0]
        # recherche d'une eventuelle css associée
        if relay_state:
            service_filter = self.manager._check_filter(relay_state)
            if service_filter not in ['default', '']:
                if os.path.exists(os.path.join('interface', '%s.css' % service_filter)):
                    css = service_filter
        try:
            css = request.args['css'][0]
        except KeyError:
            pass
        try:
            ###########################
            # traitement de l'assertion
            if authenticated:
                # l'utilisateur est déjà authentifié
                pass
                # XXX FIXME vérifier que l'utilisateur de l'assertion correspond ?
            if 'SAMLResponse' not in request.args:
                # pas d'authentification détectée
                raise InternalError(_('SAMLResponse expected'))
            else:
                # Traitement de la requête reçue (AuthnStatement)
                success, infos =  process_assertion(self.manager, request)

        except InternalError, e:
            traceback.print_exc()
            return gen_page_err(str(e), relay_state, css)
        # l'assertion reçue est valide, on la traite
        if not success:
            content = """Invalid Assertion received : %s""" % infos
            return gen_page_err(content, relay_state, css)
        else:
            content = """assertions received :<br/>"""
            for assertion_id, data in infos.items():
                # si des attributs sont remontés et valides, on cherche l'utilisateur correspondant (fédération par attribut commun)
                session_index, assertion_issuer, resp_attrs, name_id, auth_instant, class_ref = data
                # On essaye de créer une session pour l'utilisateur local correspondant (ou on reprend l'existante)
                if not authenticated:
                    sso_session = None
                # on regarde si la branche de recherche de l'utilisateur peut être déduite des attributs
                # (ex : n°rne présent)
                defer_federation = self.manager.authenticate_federated_user(resp_attrs, assertion_issuer, auth_instant, class_ref, sso_session)
                return defer_federation.addBoth(self.callb_federation, relay_state, assertion_id, data, content, css, request, cookies)
        return self.redirect_after_federation((None, {}), relay_state, assertion_id, data, content, css, request, cookies)

    @trace
    def callb_federation(self, result_fed, relay_state, assertion_id, data, content, css, request, cookies):
        session_id, user_data = result_fed
        session_index, assertion_issuer, resp_attrs, name_id, auth_instant, class_ref = data
        if session_id and not isinstance(session_id, failure.Failure):
            # Mise en place de cookie SSO pour la session locale
            cookies = []
            for cookie in getCookies(request):
                if cookie.name != 'EoleSSOServer':
                    cookies.append(cookie)
            cookies.append(TwCookie("EoleSSOServer", session_id, path = "/", discard=True, secure=True))
            # Si une session existe pour ce fournisseur de service, on le supprime
            self.remove_old_ticket(session_id, assertion_issuer)
            # On stocke les informations de session dans un ticket fictif (pour permettre le logout)
            sp_meta = self.manager.get_metadata(assertion_issuer)
            # recherche d'un éventuel index pour le choix du binding
            endpoint_index = None
            if 'index' in request.args:
                endpoint_index = request.args['index'][0]
            try:
                binding, service_url, response_url = get_endpoint(sp_meta, 'SingleLogoutService',
                                                                  ent_type='IDPSSODescriptor',
                                                                  index=endpoint_index)
            except InternalError:
                # XXX FIXME
                err_msg = _('no usable endpoint for %s') % assertion_issuer
                log.msg(err_msg)
                return http.Response(stream = gen_page(_('Assertion consuming error'), err_msg, css))
                pass
            app_ticket = self.manager.get_app_ticket(session_id, response_url)
            if app_ticket:
                # pour les applications saml, on utilise le filtre saml par défaut (pour avoir au moins l'attribut de federation)
                if self.manager.app_sessions[app_ticket].filter == 'default':
                    self.manager.app_sessions[app_ticket].filter = 'saml'
                # application de la css correspondante si existante
                service_filter = self.manager.app_sessions[app_ticket].filter
                if service_filter not in ['default', '']:
                    if os.path.exists(os.path.join('interface', '%s.css' % service_filter)):
                        css = service_filter
                # information supplémentaires à conserver
                self.manager.app_sessions[app_ticket].sp_ident = assertion_issuer
                self.manager.app_sessions[app_ticket].response_id = assertion_id
                self.manager.app_sessions[app_ticket].name_id = name_id
                self.manager.app_sessions[app_ticket].idp_session_index = session_index
                self.manager.sp_sessions[session_index] = app_ticket
            # XXX FIXME : Debug
            content += """assertion_id = %s<br/><hr/>
idp session_index = %s<br/>
attributes = %s<br/>
identity provider = %s<br/><hr/>
local session created<br/>""" % (assertion_id, session_index, str(resp_attrs), assertion_issuer)
            content += """%s: %s<br/>""" % (_('Redirecting to '), relay_state)
        else:
            if isinstance(session_id, failure.Failure):
                log.msg('Failure searching federated_user : %s' % session_id.getErrorMessage())
            content = _('Federation : No local user does match specified attributes')
        return self.redirect_after_federation(relay_state, content, css, cookies)

    @trace
    def redirect_after_federation(self, relay_state, content, css, cookies):
        if relay_state:
            resp = http.RedirectResponse(relay_state)
        else:
            resp = http.Response(stream = gen_page('Request processed', content, css))
        resp.headers.setHeader('Set-Cookie', cookies)
        return resp

class SamlResponse(SamlResource):
    """ressource principale
    traitement de l'envoi d'assertions SAML
    (Fournisseur d'identité)
    """

    def __init__(self, manager):
        self.manager = manager
        super(SamlResponse, self).__init__()

    @trace
    def remove_old_ticket(self, session_id, sp_ident):
        """ deletes any existing session for a given service provider
        """
        for ticket in self.manager.user_app_tickets[session_id]:
            if getattr(ticket, 'sp_ident', '') == sp_ident:
                old_ticket = ticket.ticket
                # supression du ticket des logouts en cours si il en fait partie
                if session_id in self.manager.pending_logout:
                    if old_ticket in self.manager.pending_logout[session_id][1]:
                        self.manager.pending_logout[session_id][1].remove(old_ticket)
                # suppression du ticket de la session utilisateur
                self.manager.user_app_tickets[session_id].remove(ticket)
                # suppression du ticket d'application
                del(self.manager.app_sessions[old_ticket])
                del(ticket)
                break

    def _render(self, request):
        """Génère une réponse SAML2 correspondant au profil HTTP-POST

        - verifie qu'une session SSO est déjà en cours
        - sinon, essaie d'en établir une
        - construit une Reponse SAML (contenant une assertion ou non) et l'envoie au fournisseur de service (entityID)
        """
        app_ticket = None
        relay_state = None
        app_ticket = None
        req_id = None
        sp_ident = None
        passive = False
        force_auth = False
        requested_contexts = []
        comparison = ""
        login_ticket = None
        css = DEFAULT_CSS
        if 'ticket' in request.args:
            app_ticket = request.args['ticket'][0]
        # recherche d'un éventuel index pour le choix du binding
        endpoint_index = None
        if 'index' in request.args:
            endpoint_index = request.args['index'][0]
        if 'SAMLRequest' in request.args:
            try:
                # requête SAML AuthnRequest reçue
                request_data = process_auth_request(self.manager, request)
                # stockage des données de la requête pour gérer la réponse (on génère un ticket 'LT' avec les données associées)
                login_ticket = self.manager.get_login_ticket(request_data)
                sp_ident, response_url, binding, req_id, passive, force_auth, comparison, requested_contexts = process_auth_request(self.manager, request)
                log.msg(_('SAMLRequest (%s) received from %s') % ('AuthnRequest', sp_ident))
                sp_meta = self.manager.get_metadata(sp_ident)
                # on vérifie que l'on est capable de gérer le niveau de sécurité demandé (requested_context)
                ctx_ok, required_ctx = check_required_contexts(comparison, requested_contexts)
                if not ctx_ok:
                    # XXX FIXME TODO: generate a saml negative saml response or display local error page (invalid context) ?
                    return http.Response(stream = gen_page(_('Unreachable resource'), required_ctx, css))
            except:
                traceback.print_exc()
                return http.Response(stream = gen_page(_('Eole-SSO : Error'), 'AuthnRequest: %s' % _('Error: unable to parse request'), css))
        else:
            # Authentification initiée par l'IDP
            if 'sp_ident' in request.args:
                sp_ident = request.args['sp_ident'][0]
                # recherche des metadonnées du fournisseur de services
                sp_meta = self.manager.get_metadata(sp_ident)
                if 'entityID' in sp_meta:
                    sp_ident = sp_meta['entityID']
                # recherche de l'adresse à laquelle enoyer les assertions
                try:
                    binding, service_url, response_url = get_endpoint(sp_meta, 'AssertionConsumerService', index=endpoint_index)
                except InternalError:
                    traceback.print_exc()
                    err_msg = _('no usable endpoint for %s') % sp_ident
                    return http.Response(stream = gen_page(_('Unreachable resource'), err_msg, css))
            else:
                # XXX FIXME TODO: generate a saml negative saml response or display local error page (invalid issuer) ?
                err_msg = _("Missing parameter")
                return http.Response(stream = gen_page(_('Unreachable resource'), err_msg, css))

        # verification de l'authentification
        authenticated = False
        session_id = None
        user_id = ""
        user_data = {}
        response_id = gen_random_id('_')
        cookies = getCookies(request)
        from_credentials = False
        for cookie in cookies:
            if cookie.name == 'EoleSSOServer':
                session_id = cookie.value
                # vérification de la validité de la session
                authenticated = self.manager.validate_session(session_id)
        # cas ou une (ré)authentification auprès de l'utilisateur est nécessaire
        if not authenticated or (force_auth and not app_ticket):
            # mode passif : on continue sans authentifier (réponse négative)
            if not passive:
                redirect_args = []
                for argname, argvalue in request.args.items():
                    redirect_args.append((argname, argvalue[0]))
                req_args = [('service', "%s/saml?%s" % (AUTH_FORM_URL, urllib.urlencode(redirect_args)))]
                if force_auth:
                    req_args.append(('renew', 'true'))
                if login_ticket:
                    req_args.append(('lt', login_ticket))
                redirect_url = '%s?%s' % (AUTH_FORM_URL, urllib.urlencode(req_args))
                resp = http.RedirectResponse(redirect_url)
                return resp

        # vérification des paramètres passés dans la requête
        try:
            css = request.args['css'][0]
        except KeyError:
            css = DEFAULT_CSS
        if 'RelayState' in request.args:
            relay_state = request.args['RelayState'][0]
        # La session est simulée à travers un ticket CAS factice
        # on utilise le paramètre relay_state pour permettre un filtrage sur le service de destination
        # si relay_state n'est pas spécifié ou n'est pas une url, on utilise
        # l'url de traitement des assertions comme service par défaut
        if app_ticket:
            # On regarde si le ticket a été délivré au moment de l'authentification
            ticket = self.manager.app_sessions[app_ticket]
            if ticket.session_id == session_id and \
               ticket.from_credentials == True and \
               ticket.service_url == "%s/saml" % (AUTH_FORM_URL):
                # modification du service du ticket pour correspondre à relay_state (ou consumer response_url par défaut)
                ticket.service_url = get_service_from_url(response_url)
                if sp_ident in self.manager.apps['sp_ident']:
                    ticket.filter = self.manager.apps['sp_ident'][sp_ident]
                else:
                    ticket.filter = self.manager._check_filter(ticket.service_url)
            else:
                app_ticket = None
        if not app_ticket and authenticated:
            # Si une session existe déjà pour ce fournisseur de service, on la supprime
            self.remove_old_ticket(session_id, sp_ident)
            # on génère un nouveau ticket pour ce fournisseur de service
            app_ticket = self.manager.get_app_ticket(session_id, response_url)
        if app_ticket:
            # pour les applications saml, on utilise le filtre saml par défaut (pour avoir au moins l'attribut de federation)
            if self.manager.app_sessions[app_ticket].filter == 'default':
                self.manager.app_sessions[app_ticket].filter = 'saml'
            # application de la css correspondante si existante
            service_filter = self.manager.app_sessions[app_ticket].filter
            if service_filter not in ['default', '']:
                if os.path.exists(os.path.join('interface', '%s.css' % service_filter)):
                    css = service_filter

            # récupération des données disponibles pour le service à atteindre
            app_ok, user_data = self.manager.get_user_details(app_ticket, response_url, renew=False, sections=True, keep_valid=True)
            data, filter_data = user_data
            user_data = {}
            # modification du nom des attributs en fonction du filtre
            for attrs in filter_data.values():
                for name, label in attrs.items():
                    user_data[name] = data[label]
            # Génrération d'un attribut nameid aléatoire pour cette session
            # Le fournisseur de service peut spécifier les attributs requis dans leurs métadonnées
            # (AttributeConsumingService/RequestedAttribute dans le xml)
            # Tous les attributs définis dans le filtre d'appplication seront envoyés
            user_id = gen_random_id('NAMEID_')
            # Stockage d'informations supplémentaires dans la ticket d'application
            # sp identifier
            # response id
            # id d'authnRequest si present ?
            self.manager.app_sessions[app_ticket].sp_ident = sp_ident
            self.manager.app_sessions[app_ticket].response_id = response_id
            self.manager.app_sessions[app_ticket].name_id = user_id
            from_credentials = self.manager.app_sessions[app_ticket].from_credentials

        attr_table = []
        for attr, val in user_data.items():
            if type(val) == list:
                val = ', '.join(val)
            if val == "":
                val = "&nbsp;"
            attr_table.append('<tr><td>%s</td><td>%s</td></tr>' % (escape(attr), escape(val)))
        attr_msg = _('Following attributes will be sent')
        # récupération de la date d'authentification
        auth_instant = self.manager.get_auth_instant(app_ticket)
        # on récupère le type d'authentification de la session SSO
        auth_class = self.manager.get_auth_class(app_ticket)
        # récupération de l'adresse du client effectuant la demande
        addr_client = request.chanRequest.getRemoteHost().host
        try:
            dns_client = socket.gethostbyaddr(addr_client)[0]
        except:
            # pas de résolution dns possible
            dns_client = None
        sign_method, response = saml_response.gen_response(response_id, authenticated, auth_instant, \
                                app_ticket, from_credentials, user_id, user_data, req_id, response_url, \
                                IDP_IDENTITY, sp_ident, CERTFILE, auth_class, addr_client, dns_client)
        #if request.args.has_key('show'):
        #    return http.Response(stream = response)
        if DISPLAY_FEDERATION == False:
            msg = _('Please wait, accessing requested service...')
            submit_input = ""
        else:
            # affichage des attributs envoyés
            msg = """<table border=1><tr><th colspan=2>%s</th></tr>%s</table>""" % (attr_msg, '\n'.join(attr_table))
            submit_input = """<input class="btn" type="Submit" value="%s">""" % _('Proceed to service')
        if binding == samlp.BINDING_HTTP_REDIRECT:
            # REDIRECT response
            response = encode_request(response, True)
            req_args = sign_request(CERTFILE, sign_method, 'SAMLResponse', response, relay_state)
            if req_args != []:
                redirect_url = '%s?%s' % (response_url, urllib.urlencode(req_args))
            else:
                redirect_url = response_url
            resp = http.RedirectResponse(redirect_url)
        else:
            # POST response (fallback par défaut)
            response = encode_request(response)
            relay_info = ""
            if relay_state:
                relay_info = """<input type="hidden" name="RelayState" value="%s">""" % relay_state
            form = """%s<br/>
<FORM name="federate_user" action="%s" method="POST">
<input type="hidden" name="SAMLResponse" value="%s">
%s
<input type="hidden" name="SigAlg" value="%s">
%s
</FORM>""" % (msg, response_url, response, relay_info, sign_method, submit_input)
            # recherche du nom 'humain' de l'entité si disponible
            entity_name = sp_meta.get('entity_local_id', '')
            head = "%s : <br/>%s" % (_('You are being Authenticated to entity'), entity_name or sp_ident)
            if DISPLAY_FEDERATION == False:
                head = _('Accessing to service')
                form += """<script type="text/javascript">document.federate_user.submit();</script>"""
            resp = http.Response(stream = gen_page(head, form, css))

        log.msg(_('SAMLResponse (%s) sent to %s') % ('AuthnStatement', sp_ident))

        return resp

    def locateChild(self, request, segments):
        if segments and segments[0] == 'metadata':
            return SamlMetadata(self.manager), ()
        if segments and segments[0] == 'acs':
            return SamlConsumer(self.manager), ()
        if segments and segments[0] == 'logout':
            return SamlLogout(self.manager), ()
        if segments == ():
            return self, ()
        return SamlErrorPage(responsecode.NOT_FOUND, "%s : %s" % (_('Unreachable resource'),  '/'.join(segments))), ()
